/*
 Copyright 2020 Padduck, LLC
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
  	http://www.apache.org/licenses/LICENSE-2.0
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/

package services

import (
	"errors"
	"github.com/pufferpanel/pufferpanel/v2"
	"github.com/pufferpanel/pufferpanel/v2/database"
	"golang.org/x/crypto/ssh"
	"strings"
)

type DatabaseSFTPAuthorization struct {
}

func (s *DatabaseSFTPAuthorization) Validate(username, password string) (perms *ssh.Permissions, err error) {
	parts := strings.Split(username, "|")
	if len(parts) != 2 {
		return nil, errors.New("incorrect username or password")
	}

	email := parts[0]
	serverId := parts[1]

	db, err := database.GetConnection()
	if err != nil {
		return nil, pufferpanel.ErrDatabaseNotAvailable
	}

	us := &User{DB: db}
	user, err := us.GetByEmail(email)
	if user == nil || err != nil || !us.IsValidCredentials(user, password) {
		return nil, errors.New("incorrect username or password")
	}

	ss := &Permission{DB: db}
	serverPerms, err := ss.GetForUserAndServer(user.ID, &serverId)
	if err != nil {
		return nil, errors.New("incorrect username or password")
	}

	userPerms, err := ss.GetForUser(user.ID)
	if err != nil {
		return nil, errors.New("incorrect username or password")
	}

	isAdmin := false
	for _, p := range userPerms {
		if p.Admin {
			isAdmin = true
			break
		}
	}

	if !serverPerms.SFTPServer && !isAdmin {
		return nil, errors.New("incorrect username or password")
	}

	perms = &ssh.Permissions{}
	perms.Extensions = make(map[string]string)
	perms.Extensions["server_id"] = serverId
	return perms, nil
}
